import re
import shrillecho.utility.archive_maybe_delete.general as gen
from shrillecho.types.playlist_types import Playlist as Playlist_t

class Playlist:
    def __init__(self, sp,  playlist: Playlist_t = None):
        self.__playlist: Playlist_t = playlist
        self.__sp = sp

    def get_tracks(self):
        track_ids = []

        results = self.__sp.playlist_items(self.__playlist.id)
        tracks = results['pagination']

        while results['next']:
            results = self.__sp.next(results)
            tracks.extend(results['pagination'])

        for item in tracks:
            track = item['track']
            if track is None:
                break
            track_ids.append(track['id'])

        return track_ids

    def get_all_uris(self):

        """ Given a playlist return all the uris"""
    
        playlist = "next"
        track_uris = []
        offset = 0
        limit = 1
        while playlist["next"]:
            playlist = self.__sp.playlist_items(self.__playlist, offset=offset, limit=limit)
            tracks = playlist['pagination']
            if not tracks:
                break
            track_uris.extend(tracks)
            offset += limit

        return track_uris
    
    

    def get_all_isrcs(self):
        """ Given a playlist return every single isrc (if it exists i.e not local file or copyrighted track)"""
        isrcs = set()
        for track in self.playlist_pagination():
            if track is None:
                continue
            try:
                isrcs.add((gen.extract_id(track['uri']), track['external_ids']['isrc']))
            except:
                matches = re.findall(r'\b\w*local\w*\b', track['uri'])
                if not matches:
                    print(f"no isrc for this track: {track['uri']}")
                else:
                    print(f"Local file detected, ignoring")
        return isrcs

    def playlist_pagination(self):

        """ Given a playlist yield the next page of results from it"""

        offset = 0
        limit = 100
        resp = self.__sp.playlist_items(self.__playlist, limit=limit)
        for item in resp['pagination']:
            if item['track'] is None:
                print(item)
            yield item['track']
        while resp['next'] is not None:
            offset += limit
            resp = self.__sp.playlist_items(self.__playlist, limit=limit, offset=offset)
            for item in resp['pagination']:
                if item['track'] is None:
                    print(item)
                yield item['track']



import asyncio
import hashlib
from typing import List, Set

# Internal
from shrillecho.artist.spotify_artist import SpotifyArtistUtil
from shrillecho.playlist.virtual_playlist import TracksController
from shrillecho.spotify.client import SpotifyClient
from shrillecho.track.spotify_track import SpotifyTrack


# Type definitions
from shrillecho.types.artist_types import Artist
from shrillecho.types.playlist_types import PlaylistTrack, SimplifiedPlaylistObject
from shrillecho.types.track_types import Track
from shrillecho.utility.cache import cached
from shrillecho.utility.general_utility import get_id


class SpotifyPlaylist:

    # @staticmethod
    # async def get_genres_from_playlist(sp: SpotifyClient, playlist) -> List[str]:
    #     """ Get the set of genres from a playlist (unique)

    #         Args:
    #             sp (SpotifyClient): client
    #             playlist (str): playlist ID / URI / LINK
        
    #         Returns:
    #             List[str]: list of genres
    #     """
    #     tracks: List[Track] = await sp.playlist_tracks(playlist)

    #     genre_set = set()

    #     artist_ids_to_fetch = set()
    #     for track in tracks:
    #         if track.artists[0].id:
    #             artist_ids_to_fetch.add(track.artists[0].id)
               
    #     genre_set = set()

    #     playlist_artists: list[Artist] = await SpotifyArtistUtil.fetch_several_artists(sp, artist_ids=list(artist_ids_to_fetch))

    #     for artist in playlist_artists:
    #         for genre in artist.genres:
    #             genre_set.add(genre)
        
    #     return list(genre_set)


    @staticmethod
    async def write_songs_to_playlist(sp: SpotifyClient, name: str, track_list: List[Track], user: str) -> str:
        tracks = SpotifyTrack.fetch_track_ids(track_list)
        playlist = await sp.user_playlist_create(user=user, name=name, public=False)
        limit = 50
        for i in range(0, len(tracks), limit):
            await sp.playlist_add_items(playlist['id'], tracks[i:i + limit])
        return playlist
    
    @cached(cache_key_prefix="get_artist_tracks_from_playlistss", class_type=Track, expiry=3600)
    @staticmethod
    async def get_artist_tracks_from_playlist(sp: SpotifyClient, playlist: str) -> List[Track]:

        playlist_id = get_id("playlist", playlist)
        playlist_tracks: List[Track] = await sp.playlist_tracks(playlist_id=playlist_id)

        unique_artists = set()

        all_tracks: List[Track] = []

        for track in playlist_tracks:
            
            main_artist: Artist = await SpotifyArtistUtil.determine_main_artist(track)

            if main_artist.id not in unique_artists:
                print(f"fetching tracks for: {main_artist.name}")
                all_tracks.extend(await SpotifyArtistUtil.get_artist_tracks(sp, main_artist.id))

            unique_artists.add(main_artist.id)

        return all_tracks


    @staticmethod
    @cached(cache_key_prefix="like_filtered", class_type=Track, expiry=3600)
    async def removed_liked_tracks(sp, playlist_tracks: List[Track], liked_tracks: List[Track]) -> List[Track]:
        filtered_tracks: List[Track] = await SpotifyTrack.track_difference_liked(playlist_tracks=playlist_tracks, 
                                                                                liked_tracks=liked_tracks)
        for track in filtered_tracks:
            track.liked = False
        return filtered_tracks
    
    @staticmethod
    @cached(cache_key_prefix="load_playlist_liked", class_type=Track, expiry=3600)
    async def load_playlist_with_likes(sp, playlist_tracks, liked_tracks) -> List[Track]: 
    
        tasks = [SpotifyTrack.is_liked(track, liked_tracks=liked_tracks) for track in playlist_tracks]
        results = await asyncio.gather(*tasks)
        
        for track, liked in zip(playlist_tracks, results):
            track.liked = liked

        return playlist_tracks

    @staticmethod
    async def get_playlist_tracks_filtered(sp: SpotifyClient, playlist_id, get_non_liked=False) -> List[Track]:
        playlist: TracksController = await TracksController.create_from_playlist(sp, playlist_id)
        await playlist.load_likes()
        return playlist.unliked_tracks if get_non_liked else playlist.tracks


from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from fastclasses_json import dataclass_json


@dataclass_json
@dataclass
class PublisherMetadata:
    id: int
    urn: str
    artist: str
    contains_music: bool
    isrc: str
    

@dataclass_json
@dataclass
class User:
    avatar_url: str
    first_name: str
    followers_count: int
    full_name: str
    id: int
    kind: str
    last_modified: str
    last_name: str
    permalink: str
    permalink_url: str
    uri: str
    urn: str
    username: str
    verified: bool
    city: str
    country_code: str
    badges: List[bool]  # Replace Dict with List
    station_urn: str
    station_permalink: str


@dataclass_json
@dataclass
class Format:
    protocol: str
    mime_type: str


@dataclass_json
@dataclass
class Transcoding:
    url: str
    preset: str
    duration: int
    snipped: bool
    format: Format
    quality: str


@dataclass_json
@dataclass
class Media:
    transcodings: List[Transcoding]


@dataclass
class UsefulMeta:
    title: Optional[str]
    isrc: Optional[str]
    artist: Optional[str]
    username: Optional[str]
    created_at: Optional[str]
    release_date: Optional[str]
    duration: Optional[int]
    full_duration: Optional[int]
    artwork_url: Optional[str]

    @property
    def meta_list(self):
        ret = []
        if self.artist != None:
            ret += self.artist.split()
        if self.username != None:
            ret += self.username.split()
        if self.title != None:
            ret += self.title.split()
        return ret
        

@dataclass_json
@dataclass
class SoundCloudTrack:
    artwork_url: str 
    caption: Optional[str] 
    commentable: bool # NOT USEFUL
    comment_count: int # NOT USEFUL
    created_at: str 
    description: str
    downloadable: bool 
    download_count: int 
    duration: int
    full_duration: int 
    embeddable_by: str # NOT USEFUL
    genre: str
    has_downloads_left: bool # NOT USEFUL
    id: int# NOT USEFUL
    kind: str
    label_name: str
    last_modified: str
    license: str
    likes_count: int
    permalink: str
    permalink_url: str
    playback_count: int
    public: bool
    publisher_metadata: PublisherMetadata
    purchase_title: str
    purchase_url: str
    release_date: str
    reposts_count: int
    secret_token: Optional[str]
    sharing: str
    state: str
    streamable: bool
    tag_list: str
    title: str
    track_format: str
    uri: str
    urn: str
    user_id: int
    visuals: Optional[Any]
    waveform_url: str
    display_date: str
    media: Media  # Replace Dict with List
    station_urn: str
    station_permalink: str
    track_authorization: str
    monetization_model: str
    policy: str
    user: User

    @property
    def useful_meta(self):
        return UsefulMeta(
            title=self.title if self.title else None,
            isrc=self.publisher_metadata.isrc if self.publisher_metadata else None,
            artist=self.publisher_metadata.artist if self.publisher_metadata else None,
            username=self.user.username if self.user else None,
            created_at=self.created_at or None,
            release_date=self.release_date or None,
            duration=self.duration if self.duration is not None else None,
            full_duration=self.full_duration if self.full_duration is not None else None,
            artwork_url=self.artwork_url
        )
    


from typing import List, Optional

from shrillecho.spotify.client import SpotifyClient
from shrillecho.types.artist_types import Artist, FollowedArtists


class SpotifyUserUtil:

    @staticmethod
    async def collect_artist(sp: SpotifyClient, followed_artist_page: FollowedArtists, artists: List[Artist], after=None):

        if len(followed_artist_page.artists.items) == 0:
            return artists

        followed_artist_page: FollowedArtists = await sp.current_user_followed_artists(limit=50, after=after)

        artists.extend(followed_artist_page.artists.items)

        after = followed_artist_page.artists.items[-1]

        await SpotifyUserUtil.collect_artist(sp, followed_artist_page=followed_artist_page, artists=artists, after=after)


    @staticmethod
    async def get_followed_artists(sp: SpotifyClient) -> List[Artist]:

        artists: List[Artist] = []

        followed_artist_page: FollowedArtists = await sp.current_user_followed_artists(limit=50, after=None)

        while len(followed_artist_page.artists.items) != 0:
           
            after = followed_artist_page.artists.items[-1].id
            artists.extend(followed_artist_page.artists.items)
            followed_artist_page: FollowedArtists = await sp.current_user_followed_artists(limit=50, after=after)
           
        return artists


import random
from typing import Tuple
import requests


class SpotifyClientGlitch:
    __client_id = ''
    __access_token = ''
    __client_token = ''

    def __init__(self, sp_dc: str = None, sp_key: str = None):
      

        self.__headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:124.0) Gecko/20100101 Firefox/124.0',
            'Accept-Encoding': 'gzip, deflate, br',
            'Accept-language': 'en-GB,en;q=0.5',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8',
            'Origin': 'https://open.spotify.com',
            'Sec-Fetch-Dest': 'document',
            'Sec-Fetch-Mode': 'navigate',
            'Sec-Fetch-Site': 'none',
            'Sec-Fetch-User': '?1',
            'Upgrade-Insecure-Requets': '1',
            'Te': 'trailers',
            'Alt-Used': 'open.spotify.com',
            'Host': 'open.spotify.com',
            'Connection': 'keep-alive'
        }

        self.__api_root = 'https://api.spotify.com/v1'
        self.__web_root = 'https://open.spotify.com'

        self.__dc = sp_dc
        self.__key = sp_key

        self.access_token = None
        self.__session = None

        self.setup_tokens()

    def setup_tokens(self):
        self.access_token, self.__client_id = self.get_access_token()
        # self.access_token = self.get_access_token()
        self.client_token = self.get_client_token(self.__client_id)
        

    def get_access_token(self) -> Tuple[str, str]:
        with requests.session() as session:
            session.headers = self.__headers

    
            # cookies = {
            #         "sp_key": "2fcbc265-ee10-46bb-aeba-3103e6fd364a",
            #         "sp_dc": "AQBBCb1uk5chw2ky-YnNe1e5fibnlY6eR5NX9R0rMRomrFMh9o7t4gUK0ZsINCIM4qK-g8q2z0f3-M2TA6NYzyIwOUzyAgFJHmwsjRa74TFhBy7fFmAo8lmU5fhAqC_u6l4TmetSJ01OOIRghLTXzwxPlwxgU0I_",
            #     }

            cookies = {}

       
            resp = session.get(f'{self.__web_root}/get_access_token', verify=True, cookies=cookies)
            resp_json = resp.json()
           
            # return resp_json['accessToken'], resp_json['clientId']
            return resp_json['accessToken'], resp_json['clientId']

    def get_client_token(self, client_id: str):
        with requests.session() as session:
            session.headers = { 
                    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:124.0) Gecko/20100101 Firefox/124.0",
                    "Accept": "application/json"
            }

            data = {
                "client_data": {
                    "client_version": "1.2.34.773.g9d8406e5",
                    "client_id": client_id,
                    "js_sdk_data": {
                        "device_brand": "unknown",
                        "device_model": "unknown",
                        "os": "windows",
                        "os_version": "NT 10.0",
                        "device_id": "f026e7d381f4aa57c9b21681ef16c106",
                        "device_type": "computer"
                    }
                }
            }

            resp = session.post('https://clienttoken.spotify.com/v1/clienttoken', json=data)
            resp_json = resp.json()
       
            return resp_json['granted_token']['token']

    def __enter__(self):
        request = "displayproxies"
        protocol = "http"
        timeout = "10000"
        country = "GB"
        ssl = "all"
        anonymity = "all"

        proxies = requests.get(
            f'https://api.proxyscrape.com/v2/?request={request}&protocol={protocol}&timeout={timeout}&country={country}&ssl={ssl}&anonymity={anonymity}')

        http_proxy = proxies.text.split('\n')[random.randint(0,1)]

        self.__session = requests.Session()
        self.__session.headers = self.__headers
        self.__session.proxies = {
            "http": http_proxy
        }
        self.__session.headers.update({
            'Client-Token': self.__client_token,
            'Authorization': f'Bearer {self.__access_token}'
        })
        return self.__session

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__session.close()



from datetime import datetime

import spotipy
from typing import Optional
from selenium.webdriver.firefox.webdriver import WebDriver
from shrillecho.utility.archive_maybe_delete import scraper
from shrillecho.types.album_types import Album
from shrillecho.types.track_types import Track
import hashlib
import json

def get_driver():
    sh.log("Setting up driver")
    return scraper.setup_driver()


def hash_str(url: str):
    return hashlib.md5(url.encode('utf-8')).hexdigest()


def log(message: str):
    print(f"[SHRILLERSPOT]: {message}")


# Utility
def get_artist_id(sp: spotipy.Spotify, artist_name: str) -> Optional[str]:
    """ Given the name of an artist, attempt to find their ID"""

    results = sp.search(q='artist:' + artist_name, type='artist')['artists']['pagination']
    if results:
        # Currently fetch the first result.
        return results[0]['id']
    else:
        print(f"No artist found with the name: {artist_name}")
        return None

def search_query_track(sp: spotipy.Spotify, query: str):
    try:
        track_query = sp.search(query, type="track")
        album_query = sp.search(query, type="album")

        album_id = album_query["albums"]["items"][0]["uri"]
        track_id = track_query["tracks"]["items"][0]["uri"]

        album: Album = Album.from_json(json.dumps(sp.album(album_id)))
        album_track_id = album.tracks.items[0].uri

        track_from_album: Track = Track.from_json(json.dumps(sp.track(album_track_id)))
        track_from_track: Track = Track.from_json(json.dumps(sp.track(track_id)))
    

        return track_from_track.external_ids.isrc, track_from_album.external_ids.isrc, track_from_track.name, track_from_album.name, track_from_album.duration_ms, track_from_track.duration_ms
    except: 
        print("failuire")
        return "","","","","",""



    # first_result = query["tracks"]["items"][0]
    # id = first_result["uri"]
    # print(id)
    # track: Track = Track.from_json(json.dumps(sp.track(id)))
  

    # return track.name, track.artists[0].name, track.external_ids.isrc

# Utility
def get_all_saved_tracks(sp: spotipy.Spotify, user: str, unique: True) -> list[str]:
    """ Given a user id, obtain all their liked songs """

    all_saved_tracks = []
    limit = 50
    offset = 0
    while True:
        response = sp.current_user_saved_tracks(limit=limit, offset=offset)
        saved_tracks = response['pagination']
        if not saved_tracks:
            break
        all_saved_tracks.extend(saved_tracks)
        offset += limit
        print(f"Fetched {offset} tracks...")

    return all_saved_tracks


def convert_isrcs_to_uris(sp, isrcs):
    # Todo - consider the list usage and whether it may produce dupes
    tracks = []
    for isrc in isrcs:
        print(f"progress: {isrc}")
        search = sp.search(f'isrc:{isrc}', type="track", limit=1)
        for track in search['tracks']['pagination']:
            tracks.append(track['uri'])
    return tracks


def crawl_related_artists(sp, artist, collected_artists, depth=0):
    if depth == 15:
        return

    related_artists = sp.artist_related_artists(artist)

    for artist in related_artists['artists']:
        lower_artist = artist['id']
        genres = artist['genres']
        monthly_listeners = artist['followers']['total']

        if lower_artist not in collected_artists:
            if monthly_listeners < 6000:
                if "future bass" in genres:
                    collected_artists.add(lower_artist)
                    crawl_related_artists(sp, lower_artist, collected_artists, depth + 1)

    return collected_artists


def convert_uris_to_isrcs(tracks, client):
    track_data = set()
    with client as client_handle:
        for track in tracks:
            track_response = client_handle.get(f'https://api.spotify.com/v1/tracks/{track}').json()
            track_data.add(track, track_response['external_ids']['isrc'])
    return track_data


def extract_id(uri):
    return uri.split(':')[2]


# def null(tracks, client):
#     access_token, client_token = client.tokens()
#
#     with requests.session() as session:
#         add_headers(session, client_token, access_token)
#
#         session.headers.update({
#             'Client-Token': client_token,
#             'Authorization': f'Bearer {access_token}'
#         })
#
#         isrc_set = set()
#
#         i = 0
#         for track in tracks:
#             dir = './cache/track_isrcs'
#
#             check_path = os.path.join(dir, f'{track.split(":")[2]}_isrc_cache.txt')
#             if os.path.exists(check_path):
#                 with open(check_path, 'r') as read_isrc:
#                     print(f"reading old isrc: {read_isrc.read()}")
#                     isrc_set.add(read_isrc.read())
#                     continue
#
#             try:
#                 track_data = session.get(f'https://api.spotify.com/v1/tracks/{track.split(":")[2]}', verify=True)
#                 track_data_j = track_data.json()
#                 print(f"progress: {i} / {len(tracks)}")
#                 try:
#
#                     isrc = track_data_j['external_ids']['isrc']
#                     isrc_set.add(isrc)
#
#                     if not os.path.exists(dir):
#                         os.makedirs(dir)
#                     path = os.path.join(dir, f'{track.split(":")[2]}_isrc_cache.txt')
#                     with open(path, 'w') as track_isrc_cache:
#                         track_isrc_cache.write(isrc)
#                 except Exception as e:
#
#                     print(f'no isrc found for track: {track_data_j["uri"]} ')
#             except:
#                 if track_data.status_code == 502:
#                     print("502 error investigate")
#                     exit(1)
#             i += 1
#     return isrc_set


def tracks_from_album(sp, album):
    track_ids = []

    tracks = sp.album_tracks(album)
    for track in tracks['pagination']:
        name = track['name']
        if name == 'Life':
            get_track = sp.track(track['uri'], market="US")

            try:
                pass
                # print(f"isrc: {get_track['external_ids']}")
            except:
                print("no linked_from field")

    return track_ids


def print_test():
    print("shrillecho-app test")


def json_debug(json, file_name: str = "quick_debug") -> str:
    # Get the current date and time
    current_time = datetime.now()

    # Format and print the current time in 24-hour clock format
    formatted_time = current_time.strftime("%H-%M-%S")

    with open(f'../data/json_debug/{file_name}_{formatted_time}.json', 'w') as json_debug:
        json_debug.write(json)


def get_artist_id(sp, artist_name):
    results = sp.search(q='artist:' + artist_name, type='artist')['artists']['pagination']
    if results:
        return results[0]['id']
    else:
        print(f"No artist found with the name: {artist_name}")
        return 1


def get_latest_tracks(sp, artist):
    tracks = []

    releases = sp.artist_albums(artist, album_type='single,album,appears_on,compilation', country="US", limit=20,
                                offset=0)

    sorted_data = sorted(releases['pagination'], key=lambda x: datetime.strptime(x['release_date'], "%Y-%m-%d"),
                         reverse=True)
    for item in sorted_data:
        get_album = sp.album(item['uri'])
        print(get_album['copyrights'])
        # print(item)
        # print(f"name: {item['name']} external: {sp.album(item['uri'])['external_ids']}")
        tracks.extend(tracks_from_album(sp, item['uri']))
    print(len(set(tracks)))




def write_songs_to_playlist(sp, name: str, tracks):
    playlist = sp.user_playlist_create(user='alcct3q9gddtkpcyidgx1z9yf', name=name, public=False)
    limit = 50
    for i in range(0, len(tracks), limit):
        sp.playlist_add_items(playlist['id'], list(tracks)[i:i + limit])


def get_followers_from_seed(sp: spotipy.Spotify, seed_user: str, driver: WebDriver, use_cache: bool = True) -> list:
    sh.log("getting followers")
    return scraper.get_followers(driver, seed_user, use_cache)



import os.path


class Cache:

    def __init__(self, folder, reader, writer):
        self._folder = folder
        self._reader = reader
        self._writer = writer
        self._setup_folder()
        self._root = f'cache/{self._folder}'

    def _setup_folder(self) -> None:
        if not os.path.exists(f'cache/{self._folder}'):
            os.makedirs(f'cache/{self._folder}')

    def fetch(self, file, *args) -> bool:
        check_path = os.path.join(f'{self._root}', file)
        if os.path.exists(check_path):
            with open(check_path, 'r') as cache:
                self._reader(cache, *args)
            return True
        with open(check_path, 'w') as cache:
            self._writer(cache, *args)
        return False

    def delete(self, file) -> None:
        check_path = os.path.join(f'{self._root}', file)
        if os.path.exists(check_path):
            os.remove(check_path)






from typing import List

# Types
from .base_types import *
from shrillecho.types.artist_types import SimpleArtist, Artist
from .component_types import LinkedFrom


@dataclass_json
@dataclass
class SimplifiedTrack:
    """ Track object for album"""
    artists: List[SimpleArtist]
    available_markets: List[str]
    disc_number: int
    duration_ms: int
    explicit: bool
    external_urls: ExternalUrls
    href: str
    id: str
    name: str
    preview_url: Optional[str]
    track_number: int
    type: str
    uri: str
    is_local: bool
    restrictions: Optional[Restrictions] = None
    is_playable: Optional[bool] = None
    linked_from: Optional[LinkedFrom] = None


@dataclass_json
@dataclass
class AlbumTracks:
    """ https://developer.spotify.com/documentation/web-api/reference/get-an-albums-tracks tracks object """
    href: str
    limit: int
    offset: int
    total: int
    items: List[SimplifiedTrack]
    next: Optional[str] = None
    previous: Optional[str] = None



@dataclass_json
@dataclass
class Album:
    """" Get album - https://developer.spotify.com/documentation/web-api/reference/get-an-album """
    album_type: str 
    artists: List[SimpleArtist]
    available_markets: List[str] 
    external_urls: ExternalUrls 
    id: str 
    images: List[Image]
    name: str
    release_date: str
    release_date_precision: str
    type: str#
    uri: str
    genres: Optional[List[str]] = None
    restrictions: Optional[Restrictions] = None
    copyrights: Optional[List[Copyright]] = None
    external_ids: Optional[ExternalIds] = None
    label: Optional[str] = None
    popularity: Optional[int] = None
    tracks: Optional[AlbumTracks] = None
    total_tracks: Optional[int] = None
    href: Optional[str] = None

@dataclass_json
@dataclass
class SimplifiedAlbum:
    album_type: str
    total_tracks: int
    available_markets: List[str]
    external_urls: ExternalUrls
    href: str
    id: str
    images: List[Image]
    name: str
    release_date: str
    release_date_precision: str
    restrictions: dict
    type: str
    uri: str
    artists: List[SimpleArtist]
    album_group: str


@dataclass_json
@dataclass
class ArtistAlbums:
    href: str
    limit: int
    next: str
    offset: int
    previous: str
    total: int
    items: List[SimplifiedAlbum]



@dataclass_json
@dataclass
class SeveralAlbums:
    """ Get several albums - https://developer.spotify.com/documentation/web-api/reference/get-multiple-albums"""
    albums: List[Album]





import json
import requests
from shrillecho.spotify.client import SpotifyClient
from shrillecho.types.artist_types import Artist, FollowerCount
from shrillecho.utility.cache import cached
from shrillecho.utility.general_utility import get_id


class ArtistController:

    def __init__(self, sp: SpotifyClient, artist: Artist):
        self._sp: SpotifyClient = sp
        self.artist: Artist = artist

    @cached(cache_key_prefix="followers", class_type=FollowerCount, expiry=3600)
    async def followers(self) -> int:
        """ the number of followers (ML form)

            Args:
                track (Artist): The artist to get the followers of (ML form)
        
            Returns:
                int: followers (ML form)
        """
        data = {"artist": self.artist.id}
        followers_call = requests.post("http://localhost:8002/ml", json=data)
        followers = followers_call.json()
        return FollowerCount.from_json(json.dumps(followers))
    



from __future__ import annotations
from typing import List
from shrillecho.artist.spotify_artist import SpotifyArtistUtil
from shrillecho.spotify.client import SpotifyClient
from shrillecho.track.spotify_track import SpotifyTrack
from shrillecho.types.album_types import Album
from shrillecho.types.artist_types import Artist
from shrillecho.types.playlist_types import Playlist
from shrillecho.types.soundcloud_types import User
from shrillecho.types.track_types import Track


class TracksController:
    
    """
        A class to encapsulate any collection of tracks
    """

    def __init__(self, sp: SpotifyClient, tracks: List[Track] = []):
        self.tracks: List[Track] = tracks
        self.liked_tracks: List[Track] = []
        self.unliked_tracks: List[Track] = []
        self._sp = sp

    @staticmethod
    async def create_from_playlist(sp: SpotifyClient, playlist: str):
        return TracksController(sp, tracks=await sp.playlist_tracks(playlist_id=playlist))
    
    @staticmethod
    async def create_from_album(sp: SpotifyClient, album: str):
        return TracksController(sp, tracks=await sp.album_tracks(album=album))

    async def load_playlist(self, playlist_id: str):
        self.tracks = await self._sp.playlist_tracks(playlist_id=playlist_id)
    
    async def load_likes(self):
        saved_tracks: List[Track] = await self._sp.current_user_saved_tracks()

        saved_track_map = set(saved_tracks)

        for track in self.tracks:

            if track in saved_track_map:
                track.liked = True
                self.liked_tracks.append(track)
            else:
                track.liked = False 
                self.unliked_tracks.append(track)



    async def load_album(self, album_id: str):
        self.tracks = await self._sp.album_tracks(album=album_id)

    async def write_tracks(self, name: str, user: str = "me") -> str:
    
        if user == "me":
            me: User = await self._sp.me()
            user = me.id

        tracks = SpotifyTrack.fetch_track_ids(self.tracks)
        
        playlist = await self._sp.user_playlist_create(user=user, name=name, public=False)
        limit = 50
        for i in range(0, len(tracks), limit):
            await self._sp.playlist_add_items(playlist['id'], tracks[i:i + limit])
        return playlist    

    async def artists(self, all_artists=False) -> List[Artist]:
        artist_ids_to_fetch = set()
        for track in self.tracks:
            if all_artists:
                for artist in track.artists:
                    artist_ids_to_fetch.add(artist.id)
            else:
                main_artist: Artist = SpotifyArtistUtil.determine_main_artist(track)
                if main_artist.id:
                    artist_ids_to_fetch.add(track.artists[0].id)

        playlist_artists: list[Artist] = await SpotifyArtistUtil.fetch_several_artists(self._sp, 
                                                                                   artist_ids=list(artist_ids_to_fetch))
        return playlist_artists

    async def genres(self) -> List[str]:
    
        playlist_artists: List[Artist] = self.artists()

        genre_set = set()

        for artist in playlist_artists:
            for genre in artist.genres:
                genre_set.add(genre)
        
        return list(genre_set)
    
    def __sub__(self, other: TracksController):
        other_track_sets = set(other.tracks) 
        return TracksController(sp=self._sp, 
                               tracks = [track for track in self.tracks if track not in other_track_sets], 
                               details=self.details)

    def __repr__(self):
        return f"{self.tracks}"
    
    
    def __and__(self, other):
        if isinstance(other, TracksController):
            self_set = set(self.tracks)
            other_set = set(other.tracks)
            return TracksController(sp=self._sp, tracks = self_set & other_set, details=self.details)
        return NotImplemented

    def __getitem__(self, index) -> Track:
        return self.tracks[index]
    
    def __len__(self):
        return len(self.tracks)



from typing import List

import spotipy
import requests

# Internal
from shrillecho.spotify.client import SpotifyClient
from shrillecho.types.album_types import Album, SeveralAlbums
from shrillecho.types.playlist_types import Playlist, SimplifiedPlaylistObject
from shrillecho.types.track_types import SeveralTracks, Track
from shrillecho.utility.archive_maybe_delete.spotify_client import SpotifyClientGlitch

class SpotifyTrack:


    @staticmethod
    def is_duplicate_track(track_a: Track, track_b: Track):
        pass

    @staticmethod
    async def fetch_several_albums(sp: SpotifyClient, album_ids: List[str], batch_size: int = 20) -> List[Album]:
        albums: List[Album] = []
        for i in range(0, len(album_ids), batch_size):
            chunk = album_ids[i:i + batch_size]
            more_albums: SeveralAlbums = await sp.albums(chunk)
            albums.extend(more_albums.albums)
        return albums
        

    @staticmethod
    async def fetch_several_tracks(sp: SpotifyClient, track_ids: List[str], batch_size: int = 50) -> List[Track]:
        tracks: List[Track] = []
        for i in range(0, len(track_ids), batch_size):
            chunk = track_ids[i:i + batch_size]
            more_tracks: List[Track] = await sp.tracks(chunk)
            tracks.extend(more_tracks)
        return tracks

    @staticmethod
    async def fetch_all_user_public_tracks(sp: SpotifyClient, user: str) -> List[Track]:
        tracks: List[Track] = []
        playlists: List[SimplifiedPlaylistObject] = await sp.user_playlists(user=user, batch=True, chunk_size=25)
        for playlist in playlists:
            tracks.extend(await sp.playlist_tracks(playlist_id=playlist.id, batch=True, chunk_size=25))
        return tracks

    @staticmethod
    def track_difference(track_list_A: List[Track], track_list_B: List[Track]) -> List[Track]:
        """ Given track list A and B, compute A - B i.e all tracks in A minus all the tracks from B.

            Args:
              track_list_A (List[Track]): list to be removed from
              track_list_B (List[Track]): list to choose which songs to remove
        
            Returns:
                List[Track]: A-B tracks
        """
        metadata_dict: dict = {}
        local_or_removed = 0
        for track_b in track_list_B:
           
            try:
                key = (track_b.artists[0].id, track_b.name) 
            
                if key not in metadata_dict:
                    metadata_dict[key] = set()
                metadata_dict[key].add(track_b.external_ids.isrc)
            except:
                local_or_removed +=1
                continue
           
        filtered_a = []

        x = 0
 
        for track_a in track_list_A:
            try:
                key = (track_a.artists[0].id ,track_a.name ) 
                if any(track_a.external_ids.isrc in isrc for isrc in metadata_dict.get(key, set())):
                    continue 
                if key not in metadata_dict:
                    filtered_a.append(track_a)
                    continue
                x +=1
                print(f'isrc miss: {track_a.name}')
            except:
                continue
            
        return filtered_a
    
        
    @staticmethod
    def fetch_track_ids(tracks: List[Track]) -> List[str]:
        """
            Given a list of tracks return a list of the ids only
        """

        ids = []
        for track in tracks:
            if track.id:
                ids.append(track.id)
        return ids
    

    @staticmethod
    def clean_tracks(tracks: List[Track]) -> List[Track]:
        """
            Given a list of tracks remove all tracks that dont have ISRC
        """

        local_removed_copyright = 0
        cleaned_tracks: List[Track] = []
        for track in tracks:
            if track.external_ids.isrc:
                if track.artists[0].name == "heyamina":
                    
                    print(track.external_ids.isrc)
                cleaned_tracks.append(track)
            else:
                local_removed_copyright += 1

        print(f'local_removed_copyright: {local_removed_copyright}')
        return cleaned_tracks

    @staticmethod
    async def track_difference_liked(playlist_tracks: List[Track], liked_tracks: List[Track]) -> List[Track]:
        return SpotifyTrack.track_difference(playlist_tracks, liked_tracks)
    
    @staticmethod
    async def is_liked(track, liked_tracks) -> bool:
        filtered_tracks: List[Track] = SpotifyTrack.track_difference([track], liked_tracks)
        return len(filtered_tracks) == 0
    

    @staticmethod
    async def get_radio(sp: SpotifyClient, track: str) -> List[Track]:
        data = {"track": track}
        radio_playlist = requests.post("http://localhost:8002/radio", json=data)
        radio_playlist_id = radio_playlist.json()["playlist"]
        playlist_tracks: List[Track] = await sp.playlist_tracks(radio_playlist_id, batch=True, chunk_size=1)
        return playlist_tracks



from base64 import urlsafe_b64encode
import base64
from dataclasses import dataclass
from http.server import BaseHTTPRequestHandler, HTTPServer
import json
import random
import string
import threading
import webbrowser
import requests
import spotipy
import os
import urllib.parse
import httpx



@dataclass
class SpotifyAuthContext:
    client_id: str
    client_secret: str
    scope: str
    redirect_uri: str


class SpotifyAuth:
    class RequestHandler(BaseHTTPRequestHandler):
        def __init__(self, auth_instance, *args, **kwargs):
            self.auth_instance = auth_instance
            super().__init__(*args, **kwargs)
            
        def do_GET(self):
            parsed_path = urllib.parse.urlparse(self.path)
            query = urllib.parse.parse_qs(parsed_path.query)
            code = query.get('code', [None])[0]
            state = query.get('state', [None])[0]

            if state is None:
                self.send_response(400)
                self.end_headers()
                self.wfile.write(b"State mismatch error.")
                return
        
        
            auth_options = {
                'url': 'https://accounts.spotify.com/api/token',
                'data': {
                    'code': code,
                    'redirect_uri': self.auth_instance.redirect_uri,
                    'grant_type': 'authorization_code',
                },
                'headers': {
                    'Content-Type': 'application/x-www-form-urlencoded',
                    'Authorization': 'Basic ' + base64.urlsafe_b64encode(f"{self.auth_instance.client_id}:{self.auth_instance.client_secret}".encode()).decode()
                }
            }

            response = requests.post(auth_options['url'], data=auth_options['data'], headers=auth_options['headers'])

            if response.ok:
                tokens = response.json()
                with open('.shrillecho_auth_cache', 'w') as file:
                    json.dump(tokens, file)

                self.send_response(200)
                self.end_headers()
                self.wfile.write(b'<h1>Authorization successful. You can close this window.</h1>')

                threading.Thread(target=self.server.shutdown).start()  
            else:
                self.send_response(400)
                self.end_headers()
                self.wfile.write(b"Failed to retrieve the access token from Spotify.")

    def __init__(self, client_id, client_secret, redirect_uri, scope):
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.scope = scope 

    def refresh_access_token(self, refresh_token):
        
        client_creds = f"{self.client_id}:{self.client_secret}"
        client_creds_b64 = urlsafe_b64encode(client_creds.encode()).decode()
        
        refresh_url = 'https://accounts.spotify.com/api/token'
        
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": f"Basic {client_creds_b64}"
        }
        body = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token
        }
        
        response = requests.post(refresh_url, headers=headers, data=body)
        
        if response.status_code == 200:
        
            return response.json()
        else:
            return f"Failed to refresh token. Status code: {response.status_code}, Response: {response.text}", None

    def get_client_access_token(self):
        client_id = os.environ.get("SPOTIPY_CLIENT_ID")
        client_secret = os.environ.get("SPOTIPY_CLIENT_SECRET")
        client_creds = f"{client_id}:{client_secret}"
        client_creds_b64 = urlsafe_b64encode(client_creds.encode()).decode()
        
        token_url = 'https://accounts.spotify.com/api/token'

        headers = {
            "Authorization": f"Basic {client_creds_b64}"
        }
        body = {
            "grant_type": "client_credentials"
        }
        
        response = requests.post(token_url, headers=headers, data=body)
        
        if response.status_code == 200:
            token_response_data = response.json()
            access_token = token_response_data['access_token']


            with open('.shrillecho_auth_cache', 'w') as file:
                json.dump(token_response_data, file)
            return access_token
        else:
            return f"Failed to obtain access token. Status code: {response.status_code}, Response: {response.text}"
    
    def authenticate_url(self):

        def generate_random_string(length):
            """Generate a random string of specified length."""
            return ''.join(random.choices(string.ascii_letters + string.digits, k=length))
        
        state = generate_random_string(16)
   
        spotify_authorize_url = 'https://accounts.spotify.com/authorize?' + urllib.parse.urlencode({
            'response_type': 'code',
            'client_id': self.client_id,
            'scope': self.scope,
            'redirect_uri': self.redirect_uri,
            'state': state
        })

        return spotify_authorize_url
    
    def auth_spotipy_client(access_token):
        return spotipy.Spotify(auth=access_token)
    

    def get_access_token(self, code):
        token_url = "https://accounts.spotify.com/api/token"
        headers = {
            "Authorization": "Basic " + base64.b64encode(f"{self.client_id}:{self.client_secret}".encode()).decode(),
            "Content-Type": "application/x-www-form-urlencoded",
        }

        data = {
            "code": code,
            "redirect_uri": self.redirect_uri,
            "grant_type": "authorization_code",
        }
        
        with httpx.Client() as client:
            response = client.post(token_url, headers=headers, data=data)
            if response.status_code == 200:
                return response.json()
            else:
                raise Exception(f"Failed to get access token, status code: {response.status_code}")

    @staticmethod
    def run_server(auth_instance):
        handler = lambda *args, **kwargs: SpotifyAuth.RequestHandler(auth_instance, *args, **kwargs)
        with HTTPServer(('localhost', 8081), handler) as server:
            print('Starting server, use <Ctrl-C> to stop')
            server.serve_forever()

    def authenticate_local(self):
        webbrowser.open(self.authenticate_url())
        SpotifyAuth.run_server(self)

class OAuthCredentials:

    def __init__(self, auth_context: SpotifyAuthContext):

        self.auth = SpotifyAuth(auth_context.client_id, 
                                auth_context.client_secret, 
                                auth_context.redirect_uri, 
                                auth_context.scope)
    
    def get_access_token(self) -> str:
        if not os.path.exists('.shrillecho_auth_cache'):
            self.auth.authenticate_local()
        
        with open('.shrillecho_auth_cache') as f:
            auth_cache = json.loads(f.read())
            return auth_cache['access_token']

class ClientCredentials:

    def __init__(self, auth_context: SpotifyAuthContext):
        self.auth = SpotifyAuth(auth_context.client_id, 
                                auth_context.client_secret, 
                                auth_context.redirect_uri, 
                                auth_context.scope)
    
    def get_access_token(self) -> str:
        return self.auth.get_client_access_token()



from typing import List
from neo4j import GraphDatabase

from shrillecho.types.artist_types import Artist

uri = "bolt://localhost:7687"
username = "neo4j"
password = "Zxasqw12"

driver = GraphDatabase.driver(uri, auth=(username, password))

def create_and_relate_artist(main_artist, other_artists):

    def create_and_relate_artists(tx, main_artist: Artist, other_artists: List[Artist]):
    
        main_artist_query = "MERGE (main:Artist {name: $main_artist}) RETURN main"
        tx.run(main_artist_query, main_artist=main_artist.name)
        
        for artist in other_artists:
            # For each artist in the list, create or find the artist node and create a RELATED_TO relationship
            artist_query = """
            MERGE (other:Artist {name: $artist})
            WITH other
            MATCH (main:Artist {name: $main_artist})
            MERGE (main)-[:RELATED_TO]->(other)
            """
            tx.run(artist_query, main_artist=main_artist.name, artist=artist.name)


    with driver.session() as session:
        session.execute_write(create_and_relate_artists, main_artist, other_artists)  
   


import spotipy
from shrillecho.types.album_types import AlbumTracks
from shrillecho.types.playlist_types import PlaylistTracks
from typing import List
from shrillecho.types.track_types import SavedTracks
from shrillecho.utility.general_utility import *
import httpx
import math
import asyncio
import time


def chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def fetch_header():
   
    with open('.cache', 'r') as file:
        data = json.load(file)
    access_token = data.get('access_token')
    headers = {
        "Authorization": f"Bearer {access_token}"
    }

    return headers


async def async_fetch_urls(urls, chunk_size, client, tracks, sp_type):

    """
        Implement better error handling
    """
    for batch in chunks(urls, chunk_size):
        while len(batch) != 0:
            print(f"fetching batch size: {len(batch)}")
            
            responses = await asyncio.gather(*(client.get(url, headers=fetch_header()) for url in batch))
            passed_responses = []
            indices_to_remove = []
            retry = 0
            for idx, r in enumerate(responses):
                if r.status_code == 200:
                    passed_responses.append(r)
                    indices_to_remove.append(idx)
                else:
                    if r.status_code == 429:
                      
                        retry = int(r.headers.get('Retry-After'))
                    else:
                        print(r)
                        exit(1)
            if retry != 0:
                print("rate limit waiting to try again...")
                time.sleep(retry)

            batch = [url for idx, url in enumerate(batch) if idx not in indices_to_remove]
            if len(passed_responses) > 0:
                tracks.extend(sp_fetch(response.json, sp_type) for response in passed_responses)


async def fetch_async(sp: spotipy.Spotify, sp_type, *args, chunk_size=25):

    """
        Figure out the return type method
    """
    
    ep_path = ""
    ep = None

    if sp_type == PlaylistTracks:
    
        ep_path = f"playlists/{args[0]}/tracks"
        ep = sp.playlist_items
    elif sp_type == SavedTracks:
        ep_path = f"me/tracks"
        ep = sp.current_user_saved_tracks
    elif sp_type == AlbumTracks:
        ep_path = f"albums/{args[0]}/tracks"
        ep = sp.album_tracks
    elif sp_type == SavedTracks:
        ep_path = f"me/tracks"
        ep = sp.current_user_saved_tracks
    elif sp_type == UserPlaylist:
        print("user playlists")
        ep_path = f"me/playlists"
        print("current user playlists")
    
        ep = sp.current_user_playlists
    else:
        print("invalid type passed!")
        exit(1)

    async with httpx.AsyncClient() as client:
        urls = []
        limit = 50
   
        initial_item: sp_type = sp_fetch(ep, sp_type, limit=limit, offset=0, *args)
        items: List[sp_type] = [initial_item]
     
        pages = math.ceil(initial_item.total / limit)
      
        for page in range(1, pages):
            urls.append(f"https://api.spotify.com/v1/{ep_path}?limit=50&offset={limit * page}")
            
        await async_fetch_urls(urls, chunk_size, client, items, sp_type)

        """ sort this out """
        unpacked_items = []
        for chunk in items:
            for cube in chunk.items:

                if sp_type == SavedTracks:
                    unpacked_items.append(cube.track)
                else:
                    unpacked_items.append(cube)
        return unpacked_items


async def fetch_async_paginate(sp: spotipy.Spotify, sp_type, url_offset, page_limit, *args, chunk_size=25):
    
    ep_path = ""
    ep = None

    if sp_type == PlaylistTracks:
    
        ep_path = f"playlists/{args[0]}/tracks"
        ep = sp.playlist_items
    elif sp_type == SavedTracks:
        ep_path = f"me/tracks"
        ep = sp.current_user_saved_tracks
    elif sp_type == AlbumTracks:
        ep_path = f"albums/{args[0]}/tracks"
        ep = sp.album_tracks
    elif sp_type == SavedTracks:
        ep_path = f"me/tracks"
        ep = sp.current_user_saved_tracks
    elif sp_type == UserPlaylist:
       
        ep_path = f"me/playlists"

        ep = sp.current_user_playlists
    else:
        print("invalid type passed!")
        exit(1)

    async with httpx.AsyncClient() as client:
        urls = []
        limit = 50

        items: List[sp_type] = []
    
        for page in range(url_offset, url_offset+page_limit):
        
            print(f"https://api.spotify.com/v1/{ep_path}?limit=50&offset={limit * page}")
            urls.append(f"https://api.spotify.com/v1/{ep_path}?limit=50&offset={limit * page}")

        await async_fetch_urls(urls, chunk_size, client, items, sp_type)

        unpacked_items = []
        for chunk in items:
            for cube in chunk.items:
                if sp_type == SavedTracks:
                    unpacked_items.append(cube.track)
                else:
                    unpacked_items.append(cube)
        return unpacked_items


from tkinter.tix import CheckList
from typing import List, TYPE_CHECKING, Self

import strawberry
from .base_types import *
from shrillecho.types.album_types import Album
from shrillecho.types.artist_types import Artist
from shrillecho.types.component_types import LinkedFrom

import hashlib

def string_to_number(s):
    if "Among" in s:
        print(f"input_string : \"{s}\"")
    # Convert the string to bytes
    s_bytes = s.encode('utf-8')
    
    # Calculate the SHA-256 hash
    hash_object = hashlib.sha256(s_bytes)
    hash_value = hash_object.hexdigest()
    
    # Convert the hexadecimal hash value to an integer
    number = int(hash_value, 16)
    if "Among" in s:
        print(f"output: {number}")
    return number

@dataclass_json
@dataclass
class Track:
    """ Get track - https://developer.spotify.com/documentation/web-api/reference/get-track """
    album: Album
    artists: List[Artist]
    available_markets: List[str]
    disc_number: int
    duration_ms: int
    explicit: bool
    external_urls: ExternalUrls
    external_ids: ExternalIds
    href: str
    id: str
    name: str
    preview_url: str
    track_number: int
    type: str
    uri: str
    is_local: bool
    popularity: str
    restrictions: Optional[Restrictions] = None
    is_playable: Optional[bool] = None
    linked_from: Optional[LinkedFrom] = None
    liked: Optional[bool] = None

    def __eq__(self, other):
        if isinstance(other, Track):
         
            try:
                if self.external_ids.isrc == other.external_ids.isrc:
                    return True 
                elif (self.artists[0].id == other.artists[0].id) and (self.name == other.name):
                    return True 
                
              
                return False
            except:
             
                return False
        return False
    
    def __hash__(self):

        # if we have valid artist use that otherwise just use the name
        try:
            return hash((str.lower(self.name), str.lower(self.artists[0].id)))
        except:
            return hash((str.lower(self.name)))
      


@dataclass
class TrackSearchItems:
    href: str
    limit: int
    next: str
    offset: int
    previous: str
    total: int
    items: List[Track]

@dataclass_json
@dataclass
class AlbumSearchItems:
    href: str
    limit: int
    next: str
    offset: int
    previous: str
    total: int
    items: List[Album]


@dataclass_json
@dataclass
class ArtistSearchItems:
    href: str
    limit: int
    next: str
    offset: int
    previous: str
    total: int
    items: List[Artist]




@dataclass_json
@dataclass
class TrackSearch:
    tracks: TrackSearchItems

@dataclass_json
@dataclass
class AlbumSearch:
    albums: AlbumSearchItems

@dataclass_json
@dataclass
class ArtistSearch:
    artists: ArtistSearchItems


@dataclass_json
@dataclass
class SavedTrack:
    added_at: str
    track: Track


@dataclass_json
@dataclass
class SavedTracks:
    """ https://developer.spotify.com/documentation/web-api/reference/get-users-saved-tracks """
    href: str
    limit: int
    next: str
    offset: int
    previous: str
    total: int
    items: List[SavedTrack]

@dataclass_json
@dataclass
class TrackInfo:
    href: str
    total: int

@dataclass_json
@dataclass
class SeveralTracks:
    """ Get several tracks - https://developer.spotify.com/documentation/web-api/reference/get-several-tracks"""
    tracks: List[Track]

@dataclass_json
@dataclass
class ReccSeed:
    afterFilteringSize: int
    afterRelinkingSize: int
    href: str
    id: str
    initialPoolSize: int
    type: str

@dataclass_json
@dataclass
class Recc:
    """ Get several tracks - https://developer.spotify.com/documentation/web-api/reference/get-several-tracks"""
    seeds: List[ReccSeed]
    tracks: List[Track]

@dataclass_json
@dataclass
class AudioFeatures:
    acousticness: float
    analysis_url: str
    danceability: float
    duration_ms: int
    energy: float
    id: str
    instrumentalness: float
    key: int
    liveness: float
    loudness: float
    mode: int
    speechiness: float
    tempo: float
    time_signature: int
    track_href: str
    type: str
    uri: str
    valence: float


from abc import ABC, abstractmethod
import datetime
from functools import wraps
import hashlib
import json
from dataclasses import asdict
from pymongo import MongoClient
import redis



class Cache(ABC):

    @abstractmethod
    async def cache_query(self, cache_key: str, cache_update_function, expiry: int, class_type: type, **kwargs):
        pass

class Mongo(Cache):
    def __init__(self, domain: str, port: str, db: str = None):
        self._mongo_client = MongoClient(f"mongodb://{domain}:{port}/")
        self._db = self._mongo_client[db]

    def set_db(self, db: str):
        self._db = self._mongo_client[db]

    async def cache_query(self, cache_key, cache_update_function, expiry, class_type, **kwargs):
        collection = self._db[kwargs.get('col')]

        doc = collection.find_one({'_id': cache_key})

        if doc:
            print(f"using cached: {cache_key}")
           
            doc.pop('expireAt', None) 
            if 'items' in doc and isinstance(doc['items'], list):
                return [class_type.from_json(json.dumps(item)) for item in doc['items']]
            else:
                return class_type.from_json(json.dumps(doc))

        print(f"fetching new: {cache_key}")
        new_data = await cache_update_function()

        if isinstance(new_data, list):
            data_dict = {'_id': cache_key, 'items': [asdict(item) for item in new_data]}
        else:
            data_dict = asdict(new_data)
            data_dict['_id'] = cache_key
      
        data_dict['expireAt'] = datetime.datetime.now() + datetime.timedelta(seconds=expiry)

        collection.replace_one({'_id': cache_key}, data_dict, upsert=True)
        
        return new_data

class Redis(Cache):

    def __init__(self, host: str, port: str):
        self.redis_client = redis.Redis(host=host, port=port, db=0)

    async def cache_query(self, cache_key: str, cache_update_function, expiry, class_type, **kwargs):
        cached_value = self.redis_client.get(cache_key)
        if cached_value:
            retrieved_value = json.loads(cached_value)
            # print(f"[REDIS CACHE] Existing entry {cache_key}")
            if isinstance(retrieved_value, list):
                return [class_type.from_json(json.dumps(item)) for item in retrieved_value]

            return class_type.from_json(json.dumps(retrieved_value))
        else:
            # print(f"[REDIS CACHE] New entry {cache_key}")
            new_value = await cache_update_function()

            if isinstance(new_value, list):
        
                new_value_dict = []

                for item in new_value:
                    if item:
                        new_value_dict.append(asdict(item))
            else:

                print(type(new_value))
                new_value_dict = asdict(new_value)

            self.redis_client.set(cache_key, json.dumps(new_value_dict), ex=expiry)
            
            return new_value


mongo_client = Mongo('localhost', '27017', 'shrillecho')
redis_client = Redis('redis', '6379')

def set_redis_host(host):
    redis_client = Redis(host, '6379')

def cached(cache_key_prefix, class_type, expiry=3600, invalidate_cache=False):
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
     
            args_str = "_".join([str(arg) for arg in args[1:]])
            kwargs_str = "_".join([f"{k}_{v}" for k, v in kwargs.items()])
            dynamic_part = f"{args_str}_{kwargs_str}"

            dynamic_hash = hashlib.sha256(dynamic_part.encode()).hexdigest()


            full_cache_key = f"{cache_key_prefix}_{dynamic_hash}"
        
            # if invalidate_cache:
            #     await CacheManager.invalidate_cache(full_cache_key)

            async def update_function():
                return await func(*args, **kwargs)
            
            return await redis_client.cache_query(full_cache_key, update_function, expiry, class_type)
        return wrapper
    return decorator















from typing import List
from shrillecho.playlist.virtual_playlist import TracksController
from shrillecho.spotify.client import SpotifyClient
from shrillecho.types.album_types import Album
from shrillecho.types.track_types import Track


class AlbumController:

    def __init__(self, sp: SpotifyClient ,album: Album, tracks: TracksController):

        self._sp: SpotifyClient = sp
        self.album: Album = album
        self.album_tracks: TracksController = tracks
        
    @staticmethod
    async def create(sp: SpotifyClient, album_id: str = None, album: Album = None):

        if not album:
            album = await sp.album(album_id=album_id)
        
        if not album_id:
            album_id = album.id

        tracks = await TracksController.create_from_album(album_id)

        return AlbumController(sp=sp, album=album, tracks=tracks)
    
    async def internal_tracks(self) -> List[Track]:
        return self.album_tracks.tracks


import os

from selenium import webdriver
from selenium.webdriver.firefox.service import Service
from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.webdriver.support.ui import WebDriverWait
from shrillecho.utility.archive_maybe_delete import general as sh
import re


class Scraper:
    def __init__(self):
        self.__driver = self.setup_driver()

    def get_driver(self):
        return self.__driver

    def wait_on_load(self, driver, pattern):
        page_source = driver.page_source
        matches = set(re.findall(pattern, page_source))
        return len(matches) > 1

    def fetch_radio_seed_en(self, driver, track):
        url = f'https://everynoise.com/research.cgi?mode=radio&name=spotify:track:{track}'
        driver.get(url)
        pattern = r'trackid="([\w]+)"'
        matches = set(re.findall(pattern, driver.page_source))

        tracks = []
        for match in matches:
            tracks.append(f'spotify:track:{match}')
        return tracks

    def fetch_source(self, driver, uri, pattern):
        driver.get(f'{uri}')
        WebDriverWait(driver, 10).until(lambda driver=driver: self.wait_on_load(driver, pattern))
        return driver.page_source

    def cache(self, path, mode, content=None):
        if not os.path.exists('cache'):
            os.makedirs('cache')

        with open(path, mode, encoding='utf-8') as cache:
            if mode == 'r':
                return cache.read()
            elif mode == 'w':
                cache.write(content)

    def get_followers(self, seed_user: str, use_cache=True):
        base_uri = 'https://open.spotify.com/user'
        uri = f'{base_uri}/{seed_user}/followers'
        hashed_uri = sh.hash_str(uri)
        pattern = r'/user/(\w+)'

        cache_path = os.path.join('cache', hashed_uri)

        if os.path.exists(f'cache/{hashed_uri}') and use_cache:
            sh.log(message='Using cached response...')
            return set(re.findall(pattern, self.cache(cache_path, 'r')))

        source = self.fetch_source(self.__driver, uri, pattern)

        self.cache(cache_path, 'w', content=source)

        return set(re.findall(pattern, source))

    def get_playlists(self, driver):
        pattern = r'/playlist/(\w+)'
        WebDriverWait(driver, 10).until(lambda driver=driver: re.search(pattern, driver.page_source))
        source = driver.page_source
        matches = re.findall(pattern, source)
        return matches
    
    def get_Tracks(self, driver):
        pattern = r'/track/(\w+)'
        WebDriverWait(driver, 10).until(lambda driver=driver: re.search(pattern, driver.page_source))
        source = driver.page_source
        matches = re.findall(pattern, source)
        return matches

    def setup_driver(self) -> WebDriver:
        # Define the options for Firefox, including the headless mode
        firefox_options = webdriver.FirefoxOptions()
        firefox_options.add_argument('--headless')

        service = Service()
        return webdriver.Firefox(service=service, options=firefox_options)

    def add_cookies(self, driver, sp_dc, sp_key):
        cookies = [
            {
                "name": "sp_dc",
                "value": sp_dc,
                "domain": "open.spotify.com"
            },
            {
                "name": "sp_key",
                "value": sp_key,
                "domain": "open.spotify.com"
            },
        ]

        for cookie in cookies:
            driver.add_cookie(cookie)
        driver.refresh()



#  async def _async_fetch_urls(self, urls, chunk_size, client, tracks, sp_type):
#         def chunks(lst, n):
#             for i in range(0, len(lst), n):
#                 yield lst[i:i + n]
          
    
#         if self.auth:
#             self.access_token = self.auth
#         elif self.auth_flow:
#             self.access_token = self.auth_flow.get_access_token()
#         else:
#             raise Exception(message="No authorization provided to the client")

#         print(self.access_token)
#         headers = {"Authorization" : f"Bearer { self.access_token}"}

#         for batch in chunks(urls, chunk_size):
#             while len(batch) != 0:
#                 print(f"fetching batch size: {len(batch)}")
#                 responses = await asyncio.gather(*(client.get(url, headers=headers) for url in batch))
#                 passed_responses = []
#                 indices_to_remove = []
#                 retry = 0
#                 for idx, r in enumerate(responses):
#                     if r.status_code == 200:
#                         passed_responses.append(r)
#                         indices_to_remove.append(idx)
#                         # cache the get response for this URL, for retry purposes , wipe it 
#                         # after as we have a complete caching layer, so we dont really want to have this fully caching
#                         # always but maybe consider.
#                     else:
#                         if r.status_code == 429:
#                             retry = int(r.headers.get('Retry-After'))
#                         else:
#                             self.batch_errors += 1 
#                             raise SpotifyClient.ShrillechoException("Batch Error, setting batch size to 10", error_code=r.status_code)
#                 if retry != 0:
#                     print("rate limit waiting to try again...")
#                     await asyncio.sleep(retry)

#                 batch = [url for idx, url in enumerate(batch) if idx not in indices_to_remove]
#                 if len(passed_responses) > 0:
#                     tracks.extend(sp_fetch(response.json, sp_type) for response in passed_responses)
            
#     # add query params via dict
#     async def _batch_request(self, ep_path: str, initial_item , sp_type: Type[T], query_params = None, override_urls = None, chunk_size=25):
#         async with httpx.AsyncClient() as client:
#             urls = []
#             limit = 50
    
#             items: List[sp_type] = [initial_item]
        
#             pages = math.ceil(initial_item.total / limit)
        
#             for page in range(1, pages):
#                 urls.append(f"https://api.spotify.com/v1/{ep_path}?limit=50&offset={limit * page}")
                
#             try:
#                 await self._async_fetch_urls(urls, chunk_size, client, items, sp_type)
#             except SpotifyClient.ShrillechoException as e:
#                  if self.batch_errors > 3:
#                       raise Exception("Fatal Error, Batch request failed after 3 retries")
#                  items: List[sp_type] = [initial_item]
#                  await self._async_fetch_urls(urls=urls, client=client, tracks=items, sp_type=sp_type, chunk_size=10)

#             self.batch_errors = 0

#             """ sort this out, essentially this batcher is designed for paginated requests
#                 so we want the items bit all collated into one sequential list
#             """
#             unpacked_items = []
        
#             for chunk in items:
#                 for cube in chunk.items:
#                         unpacked_items.append(cube)

#             return unpacked_items
        
# async def current_user_saved_tracks_b(self, chunk_size = 25, track_list=True) -> List[SavedTrack] | SavedTracks:
 
#         async def batch_request() -> List[SavedTrack]:
#             initial_item = await self._get(f"me/tracks?offset={0}&limit={50}", response_type=SavedTracks)
#             return await self._batch_request(ep_path="me/tracks", 
#                                         initial_item=initial_item,
#                                             sp_type=SavedTracks,
#                                             chunk_size=chunk_size)

#         saved_tracks: List[SavedTrack] = await redis_client.cache_query(cache_key="my_tracks_0", 
#                                           cache_update_function=batch_request, 
#                                           expiry=3600, 
#                                           class_type=SavedTrack)
#         if track_list:
#             tracks: List[Track] = []
#             for item in saved_tracks:
#                 tracks.append(item.track)
#             return tracks

#         return await saved_tracks

# async def current_user_saved_playlists(self,limit: int = None, 
#                                         offset: int = None,
#                                           batch=False,
#                                           chunk_size=25, invalidate_cache=False) -> List[SimplifiedPlaylistObject] | UserPlaylists:
        
#         cache_key = "my_playlists_new_cache"
#         ep = "me/playlists"
        
#         if not batch:
#             return await self._get(f"{ep}?offset={offset}&limit={limit}", response_type=UserPlaylists)
        
#         async def batch_request() -> List[SimplifiedPlaylistObject]:
#             initial_item = await self._get(f"{ep}?offset={0}&limit={50}", response_type=UserPlaylists)
#             return await self._batch_request(ep_path=ep, 
#                                         initial_item=initial_item,
#                                             sp_type=UserPlaylists,
#                                             chunk_size=chunk_size)
#         if invalidate_cache:
#             self.redis.delete(cache_key)

#         return await redis_client.cache_query(cache_key=cache_key, 
#                                           cache_update_function=batch_request, 
#                                           expiry=3600, 
#                                           class_type=SimplifiedPlaylistObject)

# async def playlist_tracks(self, playlist_id, limit: int = None, 
#                                         offset: int = None,
#                                           batch=True,
#                                           chunk_size=25, invalidate_cache=False, track_list=True) -> List[PlaylistTrack] | List[Track] | PlaylistTracks:
#         if not batch:
#             return await self._get(f"playlists/{playlist_id}/tracks?offset={offset}&limit={limit}", response_type=PlaylistTracks)
        
#         async def batch_request() -> List[PlaylistTrack]:
#             initial_item = await self._get(f"playlists/{playlist_id}/tracks?offset={0}&limit={50}", response_type=PlaylistTracks)
#             return await self._batch_request(ep_path=f"playlists/{playlist_id}/tracks", 
#                                         initial_item=initial_item,
#                                             sp_type=PlaylistTracks,
#                                             chunk_size=chunk_size)
#         if invalidate_cache:
#             self.redis.delete(f"playlist_tracks_{playlist_id}")

#         playlist_tracks: List[PlaylistTrack] = await redis_client.cache_query(cache_key=f"playlist_tracks_{playlist_id}", 
#                                           cache_update_function=batch_request, 
#                                           expiry=3600, 
#                                           class_type=PlaylistTrack)
#         if track_list:
#             tracks: List[Track] = []
#             for item in playlist_tracks:
#                 tracks.append(item.track)
#             return tracks
        
#         return playlist_tracks

#  async def album_tracks(self, album: str, batch=True, chunk_size=25) -> List[SimplifiedTrack]:
#         async def batch_request() -> List[SimplifiedTrack]:


#             initial_item = await self._get(f"albums/{album}/tracks?offset={0}&limit={50}", response_type=AlbumTracks)
#             return await self._batch_request(ep_path=f"albums/{album}/tracks", 
#                                         initial_item=initial_item,
#                                             sp_type=AlbumTracks,
#                                             chunk_size=chunk_size)

#         artist_albums: List[SimplifiedTrack] = await redis_client.cache_query(cache_key=f"album_tracks_{album}", 
#                                         cache_update_function=batch_request, 
#                                         expiry=3600, 
#                                         class_type=SimplifiedTrack)
#         return artist_albums

#    async def artist_albums_bat(self, artist: str, batch=True, offset=0, include_groups="album,single,appears_on,compilation", chunk_size=25) -> List[SimplifiedAlbum]:

#         if not batch:   
        
#             return await self._get(f"artists/{artist}/albums?offset={offset}&limit={50}", response_type=ArtistAlbums)
            
#         async def batch_request() -> List[SimplifiedAlbum]:

#             query_params = {}

#             query_params['include_groups'] = include_groups

#             initial_item = await self._get(f"artists/{artist}/albums?offset={0}&limit={50}", response_type=ArtistAlbums)
#             return await self._batch_request(ep_path=f"artists/{artist}/albums", 
#                                         initial_item=initial_item,
#                                             sp_type=ArtistAlbums,
#                                             chunk_size=chunk_size,
#                                             query_params=query_params)

#         artist_albums: List[SimplifiedAlbum] = await redis_client.cache_query(cache_key=f"artist_albums_{artist}", 
#                                           cache_update_function=batch_request, 
#                                           expiry=3600, 
#                                           class_type=SimplifiedAlbum)
#         return artist_albums

    # async def user_playlists(self, user:str, batch=True, chunk_size=25) -> List[SimplifiedPlaylistObject]:
    #     if not batch:
    #         pass 

    #     async def batch_request() ->  List[SimplifiedPlaylistObject]:

    #         initial_item = await self._get(f"users/{user}/playlists?offset={0}&limit={50}", response_type=UserPlaylists)
    #         return await self._batch_request(ep_path=f"users/{user}/playlists", 
    #                                     initial_item=initial_item,
    #                                         sp_type=UserPlaylists,
    #                                         chunk_size=chunk_size)

    #     artist_albums: List[SimplifiedPlaylistObject] = await redis_client.cache_query(cache_key=f"user_playlists_{user}", 
    #                                       cache_update_function=batch_request, 
    #                                       expiry=3600, 
    #                                       class_type=SimplifiedPlaylistObject)
    #     return artist_albums